@@ -33,9 +33,62 @@
 #  include <X11/XKBlib.h>
 #  include <X11/extensions/XKBstr.h>
 
-#define CONFIG_ROOT "/desktop/gnome/accesibility/keyboard"
+#define CONFIG_ROOT "/desktop/gnome/accessibility/keyboard"
+
+#ifdef DEBUG_ACCESSIBILITY
+#define d(str)		fprintf (stderr, str)
+#else
+#define d(str)		do { } while (0)
+#endif
 
 static gboolean we_are_changing_xkb_state = FALSE;
+static int xkbEventBase;
+
+static gboolean
+xkb_enabled (void)
+{
+	static gboolean initialized = 0;
+	static gboolean have_xkb = 0;
+
+	int opcode, errorBase, major, minor;
+
+	if (initialized)
+		return have_xkb;
+
+	gdk_error_trap_push ();
+	have_xkb = XkbQueryExtension (GDK_DISPLAY (),
+				      &opcode, &xkbEventBase, &errorBase, &major, &minor)
+		&& XkbUseExtension (GDK_DISPLAY (), &major, &minor);
+	XSync (GDK_DISPLAY (), FALSE);
+	gdk_error_trap_pop ();
+
+	return have_xkb;
+}
+
+static XkbDescRec *
+get_xkb_desc_rec (void)
+{
+	XkbDescRec	*desc;
+	Status		 status;
+
+	if (!xkb_enabled ())
+		return NULL;
+
+	gdk_error_trap_push ();
+	desc = XkbGetMap (GDK_DISPLAY (), XkbAllMapComponentsMask, XkbUseCoreKbd);
+	if (desc != NULL) {
+		desc->ctrls = NULL;
+		status = XkbGetControls (GDK_DISPLAY (), XkbAllControlsMask, desc);
+	}
+	XSync (GDK_DISPLAY (), FALSE);
+	gdk_error_trap_pop ();
+
+	g_return_val_if_fail (status == Success, NULL);
+	g_return_val_if_fail (desc != NULL, NULL);
+	g_return_val_if_fail (desc->ctrls != NULL, NULL);
+
+	return desc;
+}
 
 static int
 get_int (GConfClient *client, char const *key)
@@ -69,16 +122,24 @@ set_bool (GConfClient *client, char const *key, int val)
 static void
 set_server_from_gconf (GConfEntry *ignored)
 {
-	unsigned long	 which;
-	gint32 		 enabled, enable_mask;
-	XkbDescRec	*desc;
 	GConfClient	*client = gconf_client_get_default ();
+	XkbDescRec	*desc;
+	gint32 		 enabled, enable_mask;
+	unsigned long	 which;
 
-	if (!we_are_changing_xkb_state) {
-		fprintf (stderr, "We changed gconf accessibility state\n");
+	desc = get_xkb_desc_rec ();
+	if (!desc) {
+		d ("No XKB present\n");
+		return;
+	}
+
+	if (we_are_changing_xkb_state) {
+		d ("We changed gconf accessibility state\n");
 		return;
 	} else
-		fprintf (stderr, "Someone changed gcond accessibility state\n");
+		d ("Someone changed gconf accessibility state\n");
+
+	desc->ctrls->ax_options = XkbAX_LatchToLockMask;
 
 	enable_mask =   XkbAccessXKeysMask	|
 			XkbSlowKeysMask		|
@@ -92,11 +153,6 @@ set_server_from_gconf (GConfEntry *ignored)
 
 	enabled = XkbAccessXFeedbackMask;
 
-	desc = XkbGetMap(GDK_DISPLAY (), 0, XkbUseCoreKbd);
-	XkbGetControls (GDK_DISPLAY (), XkbAllControlsMask, desc);
-
-	desc->ctrls->ax_options = XkbAX_LatchToLockMask;
-
 	/* general */
 	if (gconf_client_get_bool (client, CONFIG_ROOT "/enable", NULL))
 		enabled |= XkbAccessXKeysMask;
@@ -117,14 +173,14 @@ set_server_from_gconf (GConfEntry *ignored)
 
 	/* mouse keys */
 	if (gconf_client_get_bool (client, CONFIG_ROOT "/mousekeys_enable", NULL))
-		enabled |= XkbMouseKeysMask;
-	desc->ctrls->mk_interval     = 10;
+		enabled |= XkbMouseKeysMask | XkbMouseKeysAccelMask;
+	desc->ctrls->mk_interval     = 10;	/* msec between mousekey events */
 	desc->ctrls->mk_curve	     = 50;
-	desc->ctrls->mk_max_speed    = get_int (client,
+	desc->ctrls->mk_max_speed    = get_int (client, /* pixels / event */
 		CONFIG_ROOT "/mousekeys_max_speed");
-	desc->ctrls->mk_time_to_max  = get_int (client,
-		CONFIG_ROOT "/mousekeys_accel_time");
-	desc->ctrls->mk_delay	     = get_int (client,
+	desc->ctrls->mk_time_to_max  = get_int (client,	/* events before max */
+		CONFIG_ROOT "/mousekeys_accel_time") / desc->ctrls->mk_interval;
+	desc->ctrls->mk_delay	     = get_int (client,	/* ms before 1st event */
 		CONFIG_ROOT "/mousekeys_init_delay");
 
 	/* slow keys */
@@ -163,20 +219,25 @@ set_server_from_gconf (GConfEntry *ignored)
 		XkbMouseKeysMask | XkbMouseKeysAccelMask |
 		XkbSlowKeysMask | XkbBounceKeysMask;
 
+	gdk_error_trap_push ();
 	we_are_changing_xkb_state = TRUE;
 	XkbSetControls (GDK_DISPLAY (), which, desc);
-	XFlush (GDK_DISPLAY ());
+	XSync (GDK_DISPLAY (), FALSE);
 	we_are_changing_xkb_state = FALSE;
+	gdk_error_trap_pop ();
 }
 
 static void
 set_gconf_from_server (GConfEntry *ignored)
 {
-	XkbDescRec	*desc;
 	GConfClient	*client = gconf_client_get_default ();
+	XkbDescRec	*desc;
 
-	desc = XkbGetMap(GDK_DISPLAY (), 0, XkbUseCoreKbd);
-	XkbGetControls (GDK_DISPLAY (), XkbAllControlsMask, desc);
+	desc = get_xkb_desc_rec ();
+	if (!desc) {
+		d ("No XKB present\n");
+		return;
+	}
 
 	desc->ctrls->ax_options = XkbAX_LatchToLockMask;
 
@@ -206,8 +267,9 @@ set_gconf_from_server (GConfEntry *ignored)
 		desc->ctrls->enabled_ctrls & XkbMouseKeysMask);
 	set_int (client, CONFIG_ROOT "/mousekeys_max_speed",
 		desc->ctrls->mk_max_speed);
+	/* NOTE : mk_time_to_max is measured in events not time */
 	set_int (client, CONFIG_ROOT "/mousekeys_accel_time",
-		desc->ctrls->mk_time_to_max);
+		desc->ctrls->mk_time_to_max * desc->ctrls->mk_interval);
 	set_int (client, CONFIG_ROOT "/mousekeys_init_delay",
 		desc->ctrls->mk_delay);
 
@@ -235,20 +297,19 @@ set_gconf_from_server (GConfEntry *ignored)
 	we_are_changing_xkb_state = FALSE;
 }
 
-static int xkbEventBase;
 static GdkFilterReturn 
 cb_xkb_event_filter (GdkXEvent *xevent, GdkEvent *event, gpointer user)
 {
 	XEvent *xev = (XEvent *)xevent;
 	if (xev->xany.type == (xkbEventBase + XkbEventCode)) {
 		XkbEvent *xkbEv = (XkbEvent *) event;
-		fprintf (stderr, "xkb event\n");
+		d ("xkb event\n");
 		if(xkbEv->any.xkb_type == XkbControlsNotify) {
 			if (!we_are_changing_xkb_state) {
-				fprintf (stderr, "Someone changed XKB state\n");
+				d ("Someone changed XKB state\n");
 				set_gconf_from_server (NULL);
 			} else
-				fprintf (stderr, "We changed XKB state\n");
+				d ("We changed XKB state\n");
 		}
 		return GDK_FILTER_REMOVE;
 	}
@@ -264,15 +325,16 @@ cb_xkb_event_filter (GdkXEvent *xevent, GdkEvent *event, gpointer user)
 void
 gnome_settings_accessibility_keyboard_init (GConfClient *client)
 {
-	int opcode, errorBase, major, minor;
-
-	if (!XkbQueryExtension (GDK_DISPLAY (),
-		&opcode, &xkbEventBase, &errorBase, &major, &minor) ||
-	    !XkbUseExtension (GDK_DISPLAY (), &major, &minor))
+	if (!xkb_enabled ())
 		return;
 
+	gdk_error_trap_push ();
 	XkbSelectEvents (GDK_DISPLAY (),
 		XkbUseCoreKbd, XkbAllEventsMask, XkbAllEventsMask);
+
+	XSync (GDK_DISPLAY (), FALSE);
+	gdk_error_trap_pop ();
+
 	gdk_window_add_filter (NULL, &cb_xkb_event_filter, NULL);
 
 	gnome_settings_daemon_register_callback (CONFIG_ROOT, &set_server_from_gconf);
